﻿using PCSC.Iso7816;
using System;
using System.Collections.Generic;
using System.Text;

namespace NFC.Mifare_DESFire
{
    public class MifareDESFire
    {
        /// <summary>
        /// Create new Application with AID
        /// </summary>
        /// <param name="aid">Appilication ID</param>
        public APDUCommand CreateApplication(UInt64 aid)
        {
            throw new NotImplementedException();
        }

        public APDUCommand GetApplicationIDs()
        {
            APDUCommand cmd = new APDUCommand(IsoCase.Case2Short)
            {
                CLA = 0x90,
                INS = (byte)APDUInstructions.GET_APPLICATION_IDS
            };

            return cmd;
        }

        public UInt32[] ConvertApplicationIDs(APDUResponse response)
        {
            if(response.Body.Length % 3 != 0)
            {
                throw new Exception("Invalid Body Length.");
            }

            List<UInt32> applicationIDs = new List<UInt32>();

            for(int i = 0; i < response.Body.Length; i += 3)
            {
                UInt32 new_applicationID = 0;
                new_applicationID = (UInt32)((response.Body[i] << 16) + (response.Body[i + 1] << 8) + response.Body[i + 2]);
                applicationIDs.Add(new_applicationID);
            }

            return applicationIDs.ToArray();
        }

        /// <summary>
        /// Select Application by ID
        /// </summary>
        /// <param name="id">3 Byte ID</param>
        public APDUCommand SelectApplication(UInt32 id)
        {
            byte[] id_byte = BitConverter.GetBytes(id);

            APDUCommand cmd = new APDUCommand(IsoCase.Case4Short)
            {
                CLA = 0x90,
                INS = (byte)APDUInstructions.SELECT_APPLICATION,
                Data = new byte[]
                {
                    id_byte[0],
                    id_byte[1],
                    id_byte[2]
                },
                Le = 0x00
            };

            return cmd;
        }

        public APDUCommand Authenticate_GetChallenge(byte keyid)
        {
            APDUCommand cmd = new APDUCommand(IsoCase.Case4Short)
            {
                CLA = 0x90,
                INS = (byte)APDUInstructions.,
                Data = new byte[]
                {
                    id_byte[0],
                    id_byte[1],
                    id_byte[2]
                },
                Le = 0x00
            };
        }

        /// <summary>
        /// Delete Application by ID
        /// </summary>
        /// <param name="id">3 Byte ID</param>
        public APDUCommand DeleteApplication(UInt32 id)
        {
            byte[] id_byte = BitConverter.GetBytes(id);

            APDUCommand cmd = new APDUCommand(IsoCase.Case4Short)
            {
                CLA = 0x90,
                INS = (byte)APDUInstructions.DELETE_APPLICATION,
                Data = new byte[]
                {
                    id_byte[0],
                    id_byte[1],
                    id_byte[2]
                },
                Le = 0x00
            };

            return cmd;
        }

        /// <summary>
        /// Select Application by ID
        /// </summary>
        /// <param name="id">3 Byte ID</param>
        public APDUCommand CreateApplication(UInt32 id, byte keysetting1, byte keysetting2)
        {
            byte[] id_byte = BitConverter.GetBytes(id);

            APDUCommand cmd = new APDUCommand(IsoCase.Case4Short)
            {
                CLA = 0x90,
                INS = (byte)APDUInstructions.CREATE_APPLICATION,
                Data = new byte[]
                {
                    id_byte[0],
                    id_byte[1],
                    id_byte[2],
                    keysetting1,
                    keysetting2
                },
                Le = 0x00
            };

            return cmd;
        }

        /// <summary>
        /// Genearte KeySetting1 for Application Settings or PICC Setting
        /// </summary>
        public byte GenerateKeySetting1(ChangeApplicationKey changeKey, ChangeMasterKeySettings changeMasterKeySettings, CreateDeleteFile createDeleteFile, FileDirectoryAccess fileDirectoryAccess, ChangeMasterKey changeMasterKey)
        {
            return (byte)(((byte)changeKey << 4) | (byte)changeMasterKeySettings | (byte)createDeleteFile | (byte)fileDirectoryAccess | (byte)changeMasterKey);
        }

        /// <summary>
        /// Genearte KeySetting1 for Application Settings or PICC Setting
        /// </summary>
        /// <param name="changeKey">ID of Key for changing Application Keys</param>
        /// <returns></returns>
        public byte GenerateKeySetting1(byte changeKey, ChangeMasterKeySettings changeMasterKeySettings, CreateDeleteFile createDeleteFile, FileDirectoryAccess fileDirectoryAccess, ChangeMasterKey changeMasterKey)
        {
            if(changeKey < 0x01 || changeKey >= 0x0E)
            {
                throw new ArgumentOutOfRangeException();
            }
            return GenerateKeySetting1(changeKey, changeMasterKeySettings, createDeleteFile, fileDirectoryAccess, changeMasterKey);
        }

        /// <summary>
        /// Genearte KeySetting2 for Application Creation
        /// </summary>
        /// <param name="numberOfKeys">Number of keys that can be stored within the application (0x01-0x0D)</param>
        /// <returns></returns>
        public byte GenerateKeySetting2(CryptoOperations cryptoOperations, FileIdentifies fileIdentifies, byte numberOfKeys)
        {
            if(numberOfKeys < 0x01 || numberOfKeys >= 0x0D)
            {
                throw new ArgumentOutOfRangeException();
            }

            return (byte)((byte)cryptoOperations | (byte)fileIdentifies | numberOfKeys);
        }

        /// <summary>
        /// Crypto method of the application
        /// </summary>
        public enum CryptoOperations : byte
        {
            TDES = 0x00,
            TKTDES = 0x40,
            AES = 0x80,
        }

        /// <summary>
        /// Indicates use of 2 byte ISO/IEC 7816-4 File Identifies for files within the Application
        /// </summary>
        public enum FileIdentifies : byte
        {
            NOTUSED = 0x00,
            USED = 0x20
        }

        /// <summary>
        /// hold the Access Rights for changing application keys (Change Key command)
        /// </summary>
        public enum ChangeApplicationKey : byte
        {
            /// <summary>
            /// Application master key authentication is necessary to change any key (default)
            /// </summary>
            MASTERKEY = 0x00,
            /// <summary>
            /// Authentication with the key to be changed (same Key#) is necessary to change a key
            /// </summary>
            SAMEKEY = 0x0E,
            /// <summary>
            /// All keys (except application master key, see Bit 0) within this application are frozen
            /// </summary>
            ALLKEYS = 0x0F
        }

        /// <summary>
        /// codes whether a change of the application master key settings is allowed
        /// </summary>
        public enum ChangeMasterKeySettings : byte
        {

            /// <summary>
            /// configuration not changeable anymore (frozen)
            /// </summary>
            FROZEN = 0x00,

            /// <summary>
            /// this configuration is changeable if authenticated with the application master key (default)
            /// </summary>
            WITHMASTERKEY = 0x08
        }

        /// <summary>
        /// codes whether application master key authentication is needed before “Create File” / “Delete File”
        /// </summary>
        public enum CreateDeleteFile : byte
        {
            /// <summary>
            /// “Create File”/ “Delete File”is permitted only with application master key authentication
            /// </summary>
            ONLYMASTERKEY = 0x00,

            /// <summary>
            /// “Create File”/ “Delete File”is permitted also without application master key authentication (default)
            /// </summary>
            NOKEY = 0x04,
        }

        /// <summary>
        /// codes whether application master key authentication is needed for file directory access
        /// </summary>
        public enum FileDirectoryAccess : byte
        {
            /// <summary>
            /// Successful application master key authentication is required for executing the “Get FID List”, “Get File Settings”and “Get Key Settings”commands
            /// </summary>
            ONLYMASTERKEY = 0x00,

            /// <summary>
            /// “Get FID List”, “Get File Settings” and “Get Key Settings” commands succeed independentlyof a preceding application master key authentication (default)
            /// </summary>
            NOKEY = 0x02,
        }

        /// <summary>
        /// codes whether the application master key is changeable
        /// </summary>
        public enum ChangeMasterKey : byte
        {
            /// <summary>
            /// Application master key is not changeable anymore (frozen)
            /// </summary>
            FROZEN = 0x00,

            /// <summary>
            /// Application master key is changeable (authentication with the current application master key necessary, default)
            /// </summary>
            CHANGEABLE = 0x01,
        }
    }
}
